/* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================*/
#include <ncbi.h>
#include <blast/blast.h>

int
BlastWFContextNeighborhoodAdd(wfcp, start, len, T, process)
	BLAST_WFContextPtr	wfcp;
	size_t		start;
	size_t		len;
	BLAST_Score	T;
	int			process;
{
	BLAST_Filter	ff[2];

	ff[0].func = BlastStdWordFltr;
	ff[0].id = 1;
	ff[0].data = NULL;
	ff[0].next = &ff[1];

	ff[1].func = BlastWFContextAddWordFltr;
	ff[1].id = 2;
	ff[1].data = wfcp;
	ff[1].next = NULL;

	return BlastWFContextWordFltrApply(ff, wfcp, start, len, T, process);
}

int
BlastStdWordFltr(BLAST_FilterPtr ffp, size_t off, BLAST_LetterPtr word,
		int W, BLAST_ScoreBlkPtr sbp, BLAST_Score T, int process)
{
	int		rc;

	if ( (sbp->a1 == sbp->a2 && blast_config->exact_words)
			|| (sbp->a1 != sbp->a2 && blast_config->map_words))
		if ((rc = BlastSelfScoreWordFltr(ffp, off, word, W, sbp, T, process)) != 0)
			return rc;
	if (blast_config->neighborhood_words)
		return BlastNeighborhoodWordFltr(ffp, off, word, W, sbp, T, process);
	return 0;
}


/*
	Report words that have self-scores, s, in the range 0 < s < T.
*/
int
BlastSelfScoreWordFltr(BLAST_FilterPtr ffp, size_t off, BLAST_LetterPtr word,
		int W, BLAST_ScoreBlkPtr sbp, BLAST_Score T, int process)
{
	BLAST_Letter	nh_word[BLAST_WORDSIZE_MAX];
	BLAST_AlphaMapPtr	amp;
	BLAST_WordFltrPtr	func;
	int		rc;
	register BLAST_ScoreMat	matrix;
	register BLAST_Score	score;
	int		i;

	if (ffp == NULL || (ffp = ffp->next) == NULL || (func = ffp->func) == NULL)
		return (int)BLAST_ERR_INVAL;

	matrix = sbp->matrix;

	/* Check the scores of perfect word matches */
	if (sbp->a1 == sbp->a2) {
		for (score = 0, i = 0; i < W; ++i) {
			score += matrix[word[i]][word[i]];
		}
		if (score < T && score > 0) {
			if ((rc = func(ffp, off, word, W, score, process)) != 0)
				return rc;
		}
	}
	else
		if ((amp = sbp->amp12) != NULL) {
			for (score = 0, i = 0; i < W; ++i) {
				if (!BLAST_ALPHAMAP_TST(amp, word[i]))
					return 0;
				nh_word[i] = BlastAlphaMapChr(amp, word[i]);
				score += matrix[word[i]][nh_word[i]];
			}
			if (score < T && score > 0) {
				if ((rc = func(ffp, off, nh_word, W, score, process)) != 0)
					return rc;
			}
		}
	return 0;
}

int
BlastNeighborhoodWordFltr(BLAST_FilterPtr ffp, size_t off, BLAST_LetterPtr word,
		int W, BLAST_ScoreBlkPtr sbp, BLAST_Score T, int process)
{
	BLAST_Score	best[BLAST_WORDSIZE_MAX];
	BLAST_LetterPtr	ord[BLAST_WORDSIZE_MAX];
	size_t	ord_cnt[BLAST_WORDSIZE_MAX];
	BLAST_Score	nh_score[BLAST_WORDSIZE_MAX];
	BLAST_Letter	nh_word[BLAST_WORDSIZE_MAX];
	size_t		wm1, alphasize;
	BLAST_WordFltrPtr	func;
	int		rc;
	register BLAST_ScoreMat	matrix;
	register BLAST_Score	score;
	int		i;

	if (ffp == NULL || (ffp = ffp->next) == NULL || (func = ffp->func) == NULL)
		return (int)BLAST_ERR_INVAL;

	matrix = sbp->matrix;
	wm1 = W - 1;
	alphasize = sbp->a2->alphasize;

	nh_score[0] = 0;
	best[wm1] = T;
	for (i = wm1; i > 0; --i) {
		best[i-1] = best[i] - sbp->maxcost[word[i]];
		ord[i] = sbp->order[word[i]];
	}
	ord[0] = sbp->order[word[0]];
	ord_cnt[0] = 0;
	for (i = 0; i >= 0; --i) {
		if (ord_cnt[i] >= alphasize)
			continue;
TopOfLoop:
		score = nh_score[i]
			+ matrix[word[i]][nh_word[i] = ord[i][ord_cnt[i]++]];
		if (score < best[i])
			continue;
		if (i < wm1) {
			ord_cnt[++i] = 0;
			nh_score[i] = score;
			goto TopOfLoop;
		}
		if (score <= 0)
			continue;
		if ((rc = func(ffp, off, nh_word, W, score, process)) != 0)
			return rc;
		if (ord_cnt[i] < alphasize)
			goto TopOfLoop;
	}

	return (int)BLAST_ERR_NONE;
}


#if 0
/* this is a previous version of a neighborhood word generator */
int
BlastNWGenFoo(BLAST_WordFltrPtr func, Nlm_VoidPtr arg, size_t off, BLAST_LetterPtr word,
		int W, BLAST_ScoreBlkPtr sbp, unsigned PNTR thresholdptr, int process)
{
	BLAST_Score	best[BLAST_WORDSIZE_MAX];
	BLAST_LetterPtr	ord[BLAST_WORDSIZE_MAX];
	size_t	ord_cnt[BLAST_WORDSIZE_MAX];
	BLAST_Score	T = *thresholdptr;
	BLAST_Score	nh_score[BLAST_WORDSIZE_MAX];
	BLAST_Letter	nh_word[BLAST_WORDSIZE_MAX];
	BLAST_AlphaMapPtr	amp;
	size_t		wm1, alphasize;
	register BLAST_ScoreMat	matrix;
	register BLAST_Score	score;
	int		i;

	matrix = sbp->matrix;
	wm1 = W - 1;
	alphasize = sbp->a2->alphasize;

	/* Check the scores of perfect word matches */
	if (blast_config->exact_words && sbp->a1 == sbp->a2) {
		for (score = 0, i = 0; i < W; ++i) {
			score += matrix[word[i]][word[i]];
		}
		if (score < T && score > 0) {
			if (func(arg, off, word, W, score, process) != BLAST_ERR_NONE)
				return (int)blast_errno;
		}
	}
	else
		if (blast_config->map_words && (amp = sbp->amp12) != NULL) {
			for (score = 0, i = 0; i < W; ++i) {
				nh_word[i] = BlastAlphaMapChr(amp, word[i]);
				score += matrix[word[i]][nh_word[i]];
			}
			if (score < T && score > 0) {
				if (func(arg, off, nh_word, W, score, process) != BLAST_ERR_NONE)
					return (int)blast_errno;
			}
		}

	nh_score[0] = 0;
	best[wm1] = T;
	for (i = wm1; i > 0; --i) {
		best[i-1] = best[i] - sbp->maxcost[word[i]];
		ord[i] = sbp->order[word[i]];
	}
	ord[0] = sbp->order[word[0]];
	ord_cnt[0] = 0;
	for (i = 0; i >= 0; --i) {
TopOfLoop:
		score = nh_score[i]
			+ matrix[word[i]][nh_word[i] = ord[i][ord_cnt[i]++]];
		if (score < best[i])
			continue;
		if (i < wm1) {
			ord_cnt[++i] = 0;
			nh_score[i] = score;
			goto TopOfLoop;
		}
		if (score <= 0)
			continue;
		if (func(arg, off, nh_word, W, score, process) != BLAST_ERR_NONE)
			return (int)blast_errno;
		if (ord_cnt[i] < alphasize)
			goto TopOfLoop;
	}

	return (int)BLAST_ERR_NONE;
}
#endif /* 0 */
