/* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================*/
#include <ncbi.h>
#include <blast/blast.h>

BLAST_StrPtr
BlastStrNew(maxlen, ap, lpb)
	unsigned long	maxlen; /* max. unencoded str length */
	BLAST_AlphabetPtr	ap;
	unsigned		lpb; /* letters per byte */
{
	BLAST_StrPtr	bsp;
	BLAST_LetterPtr	str;
	BLAST_Letter	sentinel;
	size_t	alloclen;

	if (ap == NULL) {
		blast_errno = BLAST_ERR_INVAL;
		return NULL;
	}

	lpb = MAX(lpb, 1);
	if (lpb * ap->bpl > sizeof(BLAST_Letter)*CHAR_BIT) {
		blast_errno = BLAST_ERR_ENCODING;
		return NULL;
	}

	bsp = (BLAST_StrPtr)BlastCalloc(sizeof(*bsp));
	if (bsp == NULL)
		return NULL;
	bsp->mt = BLAST_MEMTYPE_MALLOC;
	if (maxlen > 0) {
		alloclen = (maxlen - 1 + lpb) / lpb;
		str = (BLAST_LetterPtr)BlastMalloc(alloclen+2);
		if (str == NULL) {
			BlastFree(bsp);
			return NULL;
		}
		if (lpb == 1)
			sentinel = ap->sentinel;
		else
			sentinel = BLAST_SOFT_SENTINEL;
		bsp->_str = str;
		bsp->str = str+1;
		bsp->str[-1] = bsp->str[0] = bsp->str[alloclen] = sentinel;
	}
	else {
		alloclen = 0;
		bsp->_str = bsp->str = NULL;
	}
	bsp->len = 0;
	bsp->fulllen = 0;
	bsp->enclen = 0;
	bsp->alloclen = alloclen;
	bsp->maxlen = maxlen;
	bsp->efflen = 0.;
	bsp->ap = ap;
	bsp->lpb = lpb;
	bsp->name = NULL;
	bsp->namelen = 0;
	return bsp;
}

int LIBCALL
BlastStrLock(bsp)
	BLAST_StrPtr	bsp;
{
	int	olock;

	if (bsp != NULL) {
		olock = bsp->locked;
		bsp->locked = TRUE;
		return olock;
	}
	return 0;
}

void
BlastStrUnlock(bsp)
	BLAST_StrPtr	bsp;
{
	if (bsp != NULL)
		bsp->locked = FALSE;
}

void
BlastStrDestruct(bsp)
	BLAST_StrPtr	bsp;
{
	BLAST_StrPtr	src;
	BLAST_StrPoolPtr	spp;
	BLAST_MemType	mt;

	if (bsp == NULL || bsp->locked)
		return;
	BlastSeqDescDestruct(bsp->descp);
	if (bsp->name != NULL && bsp->namealloc > 0) {
		BlastFree(bsp->name);
		bsp->name = NULL;
		bsp->namealloc = 0;
	}
	if (bsp->alloclen > 0 && bsp->_str != NULL) {
		BlastFree(bsp->_str);
		bsp->_str = NULL;
		bsp->alloclen = 0;
	}
	src = bsp->src;
	spp = bsp->spp;
	mt = bsp->mt;
	bsp->mt = BLAST_MEMTYPE_NONE;
	switch (mt) {
	case BLAST_MEMTYPE_MALLOC:
		BlastFree(bsp);
		break;
	case BLAST_MEMTYPE_POOL:
		PoolPut(spp, bsp);
		break;
	default:
		break;
	}
	if (src)
		BlastStrDestruct(src);
	return;
}

void
BlastStrAssign(bsp, ap, str, len)
	BLAST_StrPtr	bsp;
	BLAST_AlphabetPtr	ap;
	CharPtr	str;
	size_t	len;
{
	bsp->ap = ap;
	if (bsp->alloclen > 0)
		BlastFree(bsp->_str);
	bsp->_str = bsp->str = (BLAST_LetterPtr)str;
	bsp->len = bsp->fulllen = bsp->enclen = len;
	bsp->maxlen = 0;
	bsp->alloclen = 0;
	bsp->efflen = len;
	return;
}

BLAST_StrPtr
BlastBuf2Str(buf, buflen, from, to)
	CharPtr	buf;
	size_t	buflen;
	BLAST_AlphabetPtr	from, to;
{
	BLAST_AlphaMapPtr	amp;
	BLAST_StrPtr	sp;
	size_t	i;

	if (buflen == 0) {
		blast_errno = BLAST_ERR_INVAL;
		return NULL;
	}
	if (from != to) {
		amp = BlastAlphaMapFind(from, to);
		if (amp == NULL)
			return NULL;
	}

	sp = BlastStrNew(buflen, to, 1);
	if (sp == NULL)
		return NULL;

	if (from == to)
		MemCpy((CharPtr)sp->str, (CharPtr)buf, buflen);
	else {
		for (i = 0; i < buflen; ++i)
			sp->str[i] = amp->map[buf[i]];
	}

	sp->len = sp->enclen = buflen;
	sp->offset = 0;
	sp->fulllen = buflen;
	sp->efflen = buflen;
	sp->str[buflen] = to->sentinel;
	return sp;
}

BLAST_Error
BlastStrPack(sp, lpb)
	BLAST_StrPtr	sp;
	unsigned	lpb;
{
	register BLAST_LetterPtr	cp, lp, lpmax;
	register BLAST_Letter	ch;
	size_t	i, j;
	unsigned	bpl;

	if (sp == NULL)
		return blast_errno = BLAST_ERR_INVAL;

	lpb = MAX(lpb, 1);
	bpl = sp->ap->bpl;
	if (lpb * bpl > sizeof(BLAST_Letter)*CHAR_BIT)
		return blast_errno = BLAST_ERR_ENCODING;

	if (sp->lpb > 1)
		return blast_errno = BLAST_ERR_INVAL;

	if (sp->lpb == lpb)
		return BLAST_ERR_NONE;
	sp->lpb = lpb;

	j = (sp->len / lpb) * lpb;
	for (lp = cp = sp->str, lpmax = sp->str + j; lp < lpmax; ) {
		ch = *lp++;
		for (i = lpb; i > 1; --i) {
			ch <<= bpl;
			ch |= *lp++;
		}
		*cp++ = ch;
	}
	ch = *lp++;
	j = sp->len - j;
	for (i = 1; i < j; ++i) {
		ch <<= bpl;
		ch |= *lp++;
	}
	if (j != 0)
		*cp++ = (ch << ((lpb - j) * bpl));
	*cp = BLAST_SOFT_SENTINEL;
	sp->enclen = cp - sp->str;
	return BLAST_ERR_NONE;
}

void LIBCALL
BlastSeqDescDestruct(sdp)
	BLAST_SeqDescPtr	sdp;
{
	BLAST_SeqDescPtr	sdp2;

	while (sdp != NULL) {
		ValNodeFreeData(sdp->id);
		sdp2 = sdp->next;
		if (sdp->defline != NULL)
			BlastFree(sdp->defline);
		BlastFree(sdp);
		sdp = sdp2;
	}
}

BLAST_StrPoolPtr LIBCALL
BlastStrPoolNew(cnt)
	size_t	cnt;
{
	return (BLAST_StrPoolPtr)PoolNew(cnt, sizeof(BLAST_Str), offsetof(BLAST_Str,next), NULL);
}

BLAST_StrPtr LIBCALL
BlastStrPoolGet(p)
	BLAST_StrPoolPtr	p;
{
	BLAST_StrPtr	sp;

	sp = (BLAST_StrPtr)PoolGet(p);
	MemSet(sp, 0, sizeof(*sp));
	sp->spp = p;
	sp->mt = BLAST_MEMTYPE_POOL;
	return sp;
}

void LIBCALL
BlastStrPoolDestruct(p)
	BLAST_StrPoolPtr	p;
{
	PoolDestruct((PoolBlkPtr)p);
}
