/*
 wordfind.c

 Author:  Warren Gish
*/
#include <ncbi.h>
#include <blast/blast.h>

BLAST_WordFinderPtr
BlastWordFinderNew(int neighboring, unsigned ncontexts, BLAST_AlphabetPtr ap, unsigned lpb, size_t maxlen)
{
	BLAST_WordFinderPtr	wfp;
	BLAST_WFContextPtr	wfcp;
	unsigned	i;
	size_t	s;

	lpb = MAX(lpb, 1);
	if (lpb > sizeof(BLAST_Letter)*CHAR_BIT) {
		blast_errno = BLAST_ERR_ENCODING;
		return NULL;
	}

	/* Must have sufficient bits of encoding and integral multiple of bits */
	if (ap->bpl * lpb > CHAR_BIT*sizeof(BLAST_Letter)) {
		blast_errno = BLAST_ERR_ENCODING;
		return NULL;
	}

	if (ncontexts > BLAST_WORDFINDER_CONTEXT_MAX || ncontexts == 0) {
		blast_errno = BLAST_ERR_CONTEXT;
		return NULL;
	}

	s = sizeof(*wfp) + (ncontexts-1)*sizeof(BLAST_WFContext);
	wfp = (BLAST_WordFinderPtr) BlastCalloc(s);
	if (wfp == NULL)
		return NULL;

	wfp->state = BLAST_WFSTATE_ZOMBIE;
	wfp->previous = wfp->next = NULL;
	wfp->context_cnt = 0;
	wfp->context_max = ncontexts;
	wfp->ap = ap;
	wfp->lpb = lpb;
	wfp->bpl = ap->bpl;
	wfp->maxlen = MAX(maxlen, KBYTE);
	wfp->user.choice = 0;
	wfp->user.data.intvalue = 0;

	if (wfp->lpb == 1) {
		wfp->type = BLAST_WFTYPE_1;
		if (ap->hard_sentinel) {
			wfp->dp = dfa_new(ap->minval, ap->maxval);
			if (wfp->dp != NULL)
				dfa_halton(wfp->dp, ap->sentinel);
		}
		else
			wfp->dp = dfa_new(ap->letter_min, ap->letter_max);
	}
	else {
		wfp->type = BLAST_WFTYPE_2;
		wfp->dp = dfa_new(0, (1 << (wfp->bpl * wfp->lpb)) - 1);
		if (wfp->dp != NULL && neighboring) {
			wfp->type = BLAST_WFTYPE_1;
			dfa_xlen(wfp->dp, sizeof(BLAST_WFHashHit));
		}
	}
	if (wfp->dp == NULL) {
		BlastWordFinderDestruct(wfp);
		blast_errno = BLAST_ERR_DFA;
		return NULL;
	}
	dfa_begin(wfp->dp);

	for (i=0; i < ncontexts; ++i) {
		wfcp = &wfp->wfcontext[i];
		wfcp->wfp = wfp;
		wfcp->context = i;
	}

	wfp->state = BLAST_WFSTATE_OPEN;
	return wfp;
}


/*
	BlastWordFinderClose

	Finish up the automaton for word recognition, but do not allocate
	other potentially voluminous diagonal structures for performing searches.
	No further contexts may be defined after ...Close is called.
*/
BLAST_Error
BlastWordFinderClose(BLAST_WordFinderPtr wfp)
{
	if (wfp == NULL)
		return blast_errno = BLAST_ERR_INVAL;
	if (wfp->state >= BLAST_WFSTATE_CLOSED && wfp->state <= BLAST_WFSTATE_COMPLETE)
		return BLAST_ERR_NONE;
	if (wfp->state != BLAST_WFSTATE_OPEN)
		return blast_errno = BLAST_ERR_WFSTATE;
	if (dfa_close(wfp->dp) != dfaErrNone) {
		wfp->state = BLAST_WFSTATE_ZOMBIE;
		return blast_errno = BLAST_ERR_DFA;
	}
	wfp->state = BLAST_WFSTATE_CLOSED;
	wfp->wordfinder = BlastWordFinderSelect(wfp);
	return BLAST_ERR_NONE;
}


BLAST_Error
BlastWordFinderComplete(BLAST_WordFinderPtr wfp)
{
	if (BlastWordFinderClose(wfp) != BLAST_ERR_NONE)
		return blast_errno;
	if (BlastWordFinderDiagInit(wfp) != BLAST_ERR_NONE)
		return blast_errno;
	return BLAST_ERR_NONE;
}


BLAST_Error
BlastWordFinderDestruct(BLAST_WordFinderPtr wfp)
{
	BLAST_WordFinderPtr	wfp0, wfp2;
	size_t	s;

	if (wfp == NULL)
		return blast_errno = BLAST_ERR_INVAL;

	mproc_lock();

	if (wfp->state == BLAST_WFSTATE_INVALID) {
		mproc_unlock();
		return blast_errno = BLAST_ERR_WFSTATE;
	}

	wfp0 = wfp->previous;
	wfp2 = wfp->next;
	wfp->next = wfp->previous = NULL;
	if (wfp0 == NULL && wfp2 == NULL) {
		/* no other links to this wordfinder, so the DFA can safely be freed */
		if (wfp->dp != NULL)
			dfa_destruct(wfp->dp);
	}
	if (wfp0 != NULL)
		wfp0->next = wfp2;
	if (wfp2 != NULL)
		wfp2->previous = wfp0;

	BlastWordFinderDiagDestruct(wfp);
	s = sizeof(*wfp);
	if (wfp->context_max > 0)
		s += (wfp->context_max - 1) * sizeof(wfp->wfcontext[0]);
	Nlm_MemSet((Nlm_VoidPtr)wfp, 0, s);
	BlastFree(wfp);

	mproc_unlock();

	return BLAST_ERR_NONE;
}


BLAST_WFContextPtr
BlastWordFinderContextNew(BLAST_WordFinderPtr wfp,
		ValNodePtr user, BLAST_StrPtr query, BLAST_ScoreBlkPtr sbp,
		int W, BLAST_KarlinBlkPtr kbp)
{
	BLAST_WFContextPtr	wfcp;
	unsigned	context;

	if (wfp == NULL || query == NULL || sbp == NULL) {
		blast_errno = BLAST_ERR_INVAL;
		return NULL;
	}
	if (wfp->state != BLAST_WFSTATE_OPEN) {
		blast_errno = BLAST_ERR_WFSTATE;
		return NULL;
	}
	if (wfp->context_cnt >= wfp->context_max) {
		blast_errno = BLAST_ERR_CONTEXT;
		return NULL;
	}
	if (query->lpb != 1) {
		blast_errno = BLAST_ERR_ENCODING;
		return NULL;
	}
	if (wfp->lpb*2 - 1 > W || W > BLAST_WORDSIZE_MAX) {
		blast_errno = BLAST_ERR_WORDSIZE;
		return NULL;
	}
	if (query->len < W) {
		blast_errno = BLAST_ERR_QUERYLEN;
		return NULL;
	}
	if (sbp->a1 != query->ap || sbp->a2 != wfp->ap) {
		blast_errno = BLAST_ERR_ALPHACONFLICT;
		return NULL;
	}

	context = wfp->context_cnt++;
	wfcp = &wfp->wfcontext[context];
	if (user != NULL)
		wfcp->user = *user;
	else {
		wfcp->user.choice = 1;
		wfcp->user.data.intvalue = context;
	}
	wfcp->kbp = kbp;
	wfcp->query = query;
	wfcp->q_str = query->str;
	wfcp->q_len = query->len;
	wfcp->W = W;
	/* pragmatic decision... */
	if (W * wfp->ap->bpl > 21)
		wfcp->enclen = (W - wfp->lpb + 1) / wfp->lpb;
	else
		wfcp->enclen = (W - wfp->lpb + 1) / wfp->lpb;
	wfcp->enclet = wfcp->enclen * wfp->lpb;
	wfcp->matrix = sbp->matrix;
	wfcp->sbp = sbp;

	return wfcp;
}

int
BlastWFContextWordFltrApply(ffp, wfcp, start, len, T, slide)
	BLAST_FilterPtr	ffp;
	BLAST_WFContextPtr	wfcp;
	/* The interval over which to apply the filter */
	size_t	start,	/* starting offset */
			len;	/* segment length */
	BLAST_Score	T;
	int	slide;
{
	register BLAST_LetterPtr	str;
	register size_t	off, qlen;
	register int	rc;
	register BLAST_WFContextWordFltrFnPtr	func;

	if (wfcp == NULL || wfcp->query == NULL || wfcp->sbp == NULL)
		return blast_errno = BLAST_ERR_INVAL;

	if (ffp == NULL || (func = ffp->func) == NULL)
		return blast_errno = BLAST_ERR_INVAL;

	str = wfcp->query->str + start;
	qlen = wfcp->query->len;

	if (start >= qlen || start + len > qlen)
		return blast_errno = BLAST_ERR_DOMAIN;

	if (wfcp->W > len)
		return 0;

	for (off = start + wfcp->W, len += start; off <= len; ++off)
		if ((rc = func(ffp, off, str++, wfcp->W, wfcp->sbp, T, slide)) != 0)
			return rc;

	return 0;
}

int
BlastWFContextAddWordFltr(BLAST_FilterPtr ffp, size_t off, BLAST_LetterPtr str, int W, BLAST_Score T, int slide)
{
	BLAST_WFContextPtr	wfcp;
	BLAST_WordFinderPtr	wfp;
	DFAPtr	dp;
	DFA_PatID	pi;
	unsigned	context;

	wfcp = ffp->data;
	wfp = wfcp->wfp;
	if (wfp->state != BLAST_WFSTATE_OPEN)
		return blast_errno = BLAST_ERR_WFSTATE;
	if (wfp->type != BLAST_WFTYPE_1)
		return blast_errno = BLAST_ERR_WFTYPE;

	dp = wfp->dp;
	++wfcp->stats.neighborhood_words;
	context = wfcp->context;

	if (wfp->lpb == 1) {
		switch (wfp->context_max) {
		case 1:
			pi.u = off;
#if BLAST_LETTER_SIZE == 1
			if (dfa_add(dp, (unsigned char PNTR)str, W, pi) != NULL)
#else
			if (dfa_iadd(dp, (int PNTR)str, W, pi) != NULL)
#endif
				return BLAST_ERR_NONE;
			return (int)(blast_errno = BLAST_ERR_DFA);
			/*NOTREACHED*/
		default:
			pi.u = ((unsigned long)off << BLAST_WORDFINDER_CONTEXT_SHIFT) | context;
#if BLAST_LETTER_SIZE == 1
			if (dfa_add(dp, (unsigned char PNTR)str, W, pi) != NULL)
#else
			if (dfa_iadd(dp, (int PNTR)str, W, pi) != NULL)
#endif
				return BLAST_ERR_NONE;
			return (int)(blast_errno = BLAST_ERR_DFA);
			/*NOTREACHED*/
			break;
		}
	}

	{
		/* ?mask2 are masks used when there are 2 letters per byte */
		static BLAST_Letter lmask2[] = { 0x00, 0x0f, 0xff },
							rmask2[] = { 0x00, 0xf0, 0xff }
							;
		/* ?mask4 are masks used when there are 4 letters per byte */
		static BLAST_Letter	lmask4[] = { 0x00, 0x03, 0x0f, 0x3f, 0xff },
							rmask4[] = { 0x00, 0xc0, 0xf0, 0xfc, 0xff }
							;
		BLAST_Letter	el[(BLAST_WORDSIZE_MAX+1)/2];
		BLAST_LetterPtr	elp, str0, strf;
		BLAST_WFHashHit	hh;
		int		bpl, lpb, enclet;
		int		k, klim, left, right, n;
		register BLAST_Letter	ch;
		register BLAST_LetterPtr	cp;
		register int		i;

		str0 = cp = str;
		strf = str0 + W;
		bpl = wfp->ap->bpl;
		lpb = wfp->lpb;
		enclet = wfcp->enclet;
		str = str0;
		if (W - enclet >= lpb)
			str += lpb;
		if (slide)
			klim = lpb;
		else
			klim = 1;
		for (k = 0; k < klim; ++k, str = str0 + k) {
			for (cp = str0, ch = 0; cp < str; ++cp) {
				ch <<= bpl;
				ch |= *cp;
			}
			hh.lchr = ch;
			left = cp - str0;
			hh.lmask = (lpb == 2 ? lmask2[left] : lmask4[left]);
			elp = el;
			while (cp + lpb <= strf && (cp - str < enclet || strf - cp > lpb)) {
				ch = *cp++;
				for (i = 1; i < lpb; ++i) {
					ch <<= bpl;
					ch |= *cp++;
				}
				*elp++ = ch;
			}
			hh.enclen = elp - el + 1;
			right = strf - cp;
			hh.rmask = (lpb == 2 ? rmask2[right] : rmask4[right]);
			n = (lpb - right) * bpl;
			for (ch = 0; cp < strf; ++cp) {
				ch <<= bpl;
				ch |= *cp;
			}
			hh.rchr = (ch << n);
#if 0
printf("off=%d, lt=%d, elen=%d, rt=%d\n", off, left, hh.elen, right);
#endif

			hh.u = ((unsigned long)(off - right) << BLAST_WORDFINDER_CONTEXT_SHIFT) + context;
#if BLAST_LETTER_SIZE == 1
			if (dfa_addx(dp, (unsigned char PNTR)el, hh.enclen - 1, &hh) == NULL)
#else
			if (dfa_iaddx(dp, (int PNTR)el, hh.enclen - 1, &hh) == NULL)
#endif
				return (int)(blast_errno = BLAST_ERR_DFA);
		}
		return BLAST_ERR_NONE;
	}
	/*NOTREACHED*/
}

BLAST_WordFinderPtr LIBCALL
BlastWordFinderLink(wfp)
	BLAST_WordFinderPtr	wfp;
{
	BLAST_WordFinderPtr	nfp, wfp2;
	BLAST_WFState	wfp_state;
	size_t	s;
	int		i;

	if (wfp == NULL) {
		blast_errno = BLAST_ERR_INVAL;
		return NULL;
	}
	mproc_lock();
	wfp_state = wfp->state;
	if (wfp_state < BLAST_WFSTATE_CLOSED || wfp_state > BLAST_WFSTATE_COMPLETE) {
		mproc_unlock();
		blast_errno = BLAST_ERR_WFSTATE;
		return NULL;
	}
	if (wfp->root == NULL) {
		wfp->root = wfp;
		mproc_unlock();
		return wfp;
	}

	s = sizeof(*wfp);
	if (wfp->context_cnt > 0)
		s += (wfp->context_cnt - 1) * sizeof(wfp->wfcontext[0]);
	nfp = (BLAST_WordFinderPtr) BlastMalloc(s);
	if (nfp == NULL)
		return NULL;
	Nlm_MemCpy((Nlm_VoidPtr)nfp, (Nlm_VoidPtr)wfp, s);
	wfp2 = nfp->next = wfp->next;
	if (wfp2 != NULL)
		wfp2->previous = nfp;
	wfp->next = nfp;
	nfp->previous = wfp;
	mproc_unlock();

	nfp->context_max = nfp->context_cnt;
	for (i = 0; i < nfp->context_cnt; ++i) {
		Nlm_MemSet((Nlm_VoidPtr)&nfp->wfcontext[i].dd, 0, sizeof(nfp->wfcontext[0].dd));
		Nlm_MemSet((Nlm_VoidPtr)&nfp->wfcontext[i].stats, 0, sizeof(nfp->wfcontext[0].stats));
	}

	if (wfp_state == BLAST_WFSTATE_COMPLETE) {
		nfp->state = BLAST_WFSTATE_CLOSED;
		if (BlastWordFinderDiagInit(nfp) != BLAST_ERR_NONE) {
			BlastWordFinderDestruct(nfp);
			return NULL;
		}
	}

	return nfp;
}

BLAST_Error
BlastWordFinderUnlink(wfp)
	BLAST_WordFinderPtr	wfp;
{
	if (wfp == NULL)
		return blast_errno = BLAST_ERR_INVAL;
	mproc_lock();
	if (wfp->root == wfp || wfp->root == NULL) {
		mproc_unlock();
		return BLAST_ERR_NONE;
	}
	BlastWordFinderDestruct(wfp);
	mproc_unlock();
	return BLAST_ERR_NONE;
}


BLAST_Error
BlastWordFinderDiagInit(wfp)
	BLAST_WordFinderPtr	wfp;
{
	BLAST_WFContextPtr	wfcp;
	size_t	alloclen, qlen, slen;
	unsigned	context;

	if (wfp == NULL)
		return blast_errno = BLAST_ERR_INVAL;
	if (wfp->state != BLAST_WFSTATE_CLOSED)
		return blast_errno = BLAST_ERR_WFSTATE;

	slen = wfp->maxlen;
	for (context = 0; context < wfp->context_cnt; ++context) {
		wfcp = &wfp->wfcontext[context];
		qlen = wfcp->query->len;

		wfcp->dd.alloclen = alloclen = qlen + slen + 1 + 1;
		wfcp->dd.diag_level0 = (BLAST_OffsetPtr)BlastMalloc(alloclen*sizeof(wfcp->dd.diag_level0[0]));
		wfcp->dd.diag_stack = (BLAST_DiagPtr)BlastMalloc(alloclen*sizeof(wfcp->dd.diag_stack[0]));
		wfcp->dd.diag_level = wfcp->dd.diag_level0 + qlen;
		wfcp->dd.diag_stack_top = wfcp->dd.diag_stack;
		if (wfcp->dd.diag_level0 == NULL
				|| wfcp->dd.diag_stack == NULL) {
			BlastWordFinderDestruct(wfp);
			return blast_errno;
		}
		wfcp->dd.initlen = 0;
		/* the +1 in (qlen+1) causes the [0] element to be initialized */
		Nlm_MemSet((VoidPtr)wfcp->dd.diag_level0, 0, (qlen+1) * sizeof(wfcp->dd.diag_level0[0]));
	}

	wfp->state = BLAST_WFSTATE_COMPLETE;
	return BLAST_ERR_NONE;
}

BLAST_Error
BlastWordFinderDiagDestruct(wfp)
	BLAST_WordFinderPtr	wfp;
{
	BLAST_WFContextPtr	wfcp;
	unsigned	context;

	if (wfp == NULL)
		return blast_errno = BLAST_ERR_INVAL;
	if (wfp->state < BLAST_WFSTATE_CLOSED || wfp->state > BLAST_WFSTATE_ZOMBIE)
		return blast_errno = BLAST_ERR_WFSTATE;

	for (context = 0; context < wfp->context_cnt; ++context) {
		wfcp = &wfp->wfcontext[context];
		if (wfcp->sbp == NULL)
			continue;
		if (wfcp->dd.diag_level0 != NULL)
			BlastFree(wfcp->dd.diag_level0);
		if (wfcp->dd.diag_stack != NULL)
			BlastFree(wfcp->dd.diag_stack);
		Nlm_MemSet((Nlm_VoidPtr)&wfcp->dd, 0, sizeof(wfcp->dd));
	}

	if (wfp->state == BLAST_WFSTATE_COMPLETE)
		wfp->state = BLAST_WFSTATE_CLOSED;
	return BLAST_ERR_NONE;
}

BLAST_Error
BlastWordFinderDiagResize(wfp, maxlen)
	BLAST_WordFinderPtr	wfp;
	size_t		maxlen;
{
	if (wfp == NULL)
		return blast_errno = BLAST_ERR_INVAL;

	if (wfp->state < BLAST_WFSTATE_CLOSED || wfp->state > BLAST_WFSTATE_COMPLETE)
		return blast_errno = BLAST_ERR_WFSTATE;

	if (BlastWordFinderDiagDestruct(wfp) != BLAST_ERR_NONE)
		return blast_errno;

	wfp->maxlen = maxlen;
	wfp->state = BLAST_WFSTATE_CLOSED;
	return BlastWordFinderDiagInit(wfp);
}

BLAST_Error
BlastWordFinderSearchEnter(wfp)
	BLAST_WordFinderPtr	wfp;
{
	register BLAST_WFContextPtr	wfcp;
	register unsigned	context;
	register ssize_t	incr;

	for (context = 0; context < wfp->context_cnt; ++context) {
		wfcp = &wfp->wfcontext[context];

		wfcp->stats.word_hits = 0;
		wfcp->stats.failed_extensions = 0;
	}
	if (wfp->subject->len > wfp->wfcontext[0].dd.initlen) {
		incr = wfp->subject->len - wfp->wfcontext[0].dd.initlen;
		incr = MAX(incr, KBYTE);
		incr = MIN(wfp->maxlen - wfp->wfcontext[0].dd.initlen, incr);
		for (context = 0; context < wfp->context_cnt; ++context) {
			wfcp = &wfp->wfcontext[context];
			/* the +1 is present, because [0] element was initialized earlier */
			Nlm_MemSet((VoidPtr)&wfcp->dd.diag_level[wfcp->dd.initlen + 1], 0,
						incr * sizeof(wfcp->dd.diag_level[0]));
			wfcp->dd.initlen += incr;
		}
	}
	return BLAST_ERR_NONE;
}

BLAST_Error
BlastWordFinderSearchExit(wfp)
	BLAST_WordFinderPtr	wfp;
{
	BLAST_WFContextPtr	wfcp;
	unsigned	context;
	register BLAST_OffsetPtr	dl;
	register BLAST_DiagPtr	ds, dt;

	for (context = 0; context < wfp->context_cnt; ++context) {
		wfcp = &wfp->wfcontext[context];

		wfcp->stats.xword_hits += wfcp->stats.word_hits;
		wfcp->stats.xfailed_extensions += wfcp->stats.failed_extensions;

		dt = wfcp->dd.diag_stack_top;
		ds = wfcp->dd.diag_stack_top = wfcp->dd.diag_stack;
		dl = wfcp->dd.diag_level;
		while (dt > ds)
			dl[*--dt] = 0;
	}
	return BLAST_ERR_NONE;
}

BLAST_Error
BlastWordFinderFilterSet(wfp, ffp)
	register BLAST_WordFinderPtr	wfp;
	register BLAST_FilterPtr	ffp;
{
	register unsigned	context;

	if (wfp == NULL || ffp == NULL || ffp->func == NULL)
		return blast_errno = BLAST_ERR_INVAL;

	for (context = 0; context < wfp->context_max; ++context)
		wfp->wfcontext[context].ffp = ffp;
	return BLAST_ERR_NONE;
}

BLAST_Error
BlastWordFinderSubjectSet(wfp, subject)
	BLAST_WordFinderPtr	wfp;
	register BLAST_StrPtr	subject;
{
	BLAST_WFContextPtr	wfcp;
	unsigned	context;
	BLAST_LetterPtr	sf, sf2;
	unsigned	remain;

	if (wfp == NULL || subject == NULL)
		return blast_errno = BLAST_ERR_INVAL;

	if (subject->len > wfp->maxlen)
		return blast_errno = BLAST_ERR_SUBJECTLEN;

	wfp->subject = subject;
	sf = sf2 = subject->str + subject->enclen;
	remain = subject->len % subject->lpb;
	if (remain > 0)
		--sf2;
	for (context = 0; context < wfp->context_max; ++context) {
		wfcp = &wfp->wfcontext[context];
		wfcp->subject = subject;
		wfcp->s_str = subject->str;
		wfcp->sf = sf;
		wfcp->sf2 = sf2;
		wfcp->remain = remain;
	}
	return BLAST_ERR_NONE;
}


BLAST_Error
BlastWordFinderExtendParamSet(wfp, X, cutoff)
	register BLAST_WordFinderPtr	wfp;
	register BLAST_Score	X;
	register BLAST_Score	cutoff;
{
	register unsigned	context;

	if (wfp == NULL)
		return blast_errno = BLAST_ERR_INVAL;

	if (cutoff < 1 || X > -1)
		return blast_errno = BLAST_ERR_SCORE_DOMAIN;

	X = MAX(X, BLAST_SCORE_MIN/2);

	for (context = 0; context < wfp->context_max; ++context) {
		wfp->wfcontext[context].cutoff = cutoff;
		wfp->wfcontext[context].X = X;
	}
	return BLAST_ERR_NONE;
}

BLAST_Error
BlastWFContextExtendParamSet(wfcp, X, cutoff)
	BLAST_WFContextPtr	wfcp;
	BLAST_Score	X; /* max. drop-off score, must be negative */
	BLAST_Score	cutoff; /* cutoff score, must be positive */
{
	if (wfcp == NULL)
		return blast_errno = BLAST_ERR_INVAL;

	if (cutoff < 1 || X > -1)
		return blast_errno = BLAST_ERR_SCORE_DOMAIN;

	wfcp->cutoff = cutoff;
	wfcp->X = MAX(X, BLAST_SCORE_MIN/2);
	return BLAST_ERR_NONE;
}

BLAST_Error
BlastWFSearchStatsSum(ssp0, ssp)
	register BLAST_WFSearchStatsPtr	ssp0, ssp;
{
	if (ssp0 == NULL || ssp == NULL)
		return blast_errno = BLAST_ERR_INVAL;

	ssp0->high_score = MAX(ssp0->high_score, ssp->high_score);
	ssp0->neighborhood_words += ssp->neighborhood_words;
	ssp0->exact_words += ssp->exact_words;

	ssp0->word_hits += ssp->word_hits;
	ssp0->failed_extensions += ssp->failed_extensions;
	ssp0->xword_hits += ssp->xword_hits;
	ssp0->xfailed_extensions += ssp->xfailed_extensions;

	ssp0->spanned += ssp->spanned;
	ssp0->passed += ssp->passed;
	ssp0->reported += ssp->reported;
	ssp0->reportable += ssp->reportable;
	ssp0->saved += ssp->saved;
	return BLAST_ERR_NONE;
}
